/*
 *  Part of rvglue source code,
 *  Copyright (C) 2000 Alexander Kroeller (alilein@gmx.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include <print.h>

#include "parser.h"
#include "errdesc.h"
#include "appopts.h"

void Parser::gentoken(void)
{
  int pos;

  skipwhite();
  if( type(gchar()) == PT_TEXT )
    {
      pos=0;
      do {
	if(pos<MAX_TOK_LENGTH-1)
	  {
	    tok_text[pos]=curch;
	    pos++;
	  };
	gchar();
      } while(type(curch)==PT_TEXT);
      push();
      tok_text[pos]=0;
      curtok=PT_TEXT;
    }
  else if( type(curch) == PT_CBEGIN )
    {
      int level=1;
      int start=lineno;
      while((level > 0) && (type(curch) != PT_EOF))
	{
	  switch(type(gchar()))
	    {
	    case PT_CBEGIN:
	      level++; break;
	    case PT_CEND:
	      level--; break;
	    };
	};
      if(type(curch)==PT_EOF)
	error(ERROR_COMMENT_EOF, "Comment at line %d never closed", start);
      gentoken();
    }
  else if( type(curch) == PT_CEND )
    {
      error(ERROR_COMMENT_NOBEGIN, "Unmatched ']'");
      gentoken();
    }
  else if( type(curch) == PT_LINECOM )
    {
      do
	gchar();
      while( (type(curch)!=PT_EOF) && (curch!='\n') );
      gentoken();
    }
  else
    {
      curtok=type(curch);
    };
}

void Parser::skipwhite(void)
{
  while(type(gchar()) == PT_WHITE)
    ;

  if(type(curch) != PT_EOF)
    push();
}

int Parser::gchar(void)
{
  curch=fgetc(f);
  if(curch=='\n')
    lineno++;
  return curch;
}

void Parser::push(void)
{
  ungetc(curch, f);
}



int Parser::type(int ch)
{
  switch(ch)
    {
    case EOF:
      return PT_EOF;
    case ' ':
    case '\t':
    case '\n':
    case '\r':
      return PT_WHITE;
    case '(':
      return PT_BBEGIN;
    case ')':
      return PT_BEND;
    case '[':
      return PT_CBEGIN;
    case ']':
      return PT_CEND;
    case '#':
      return PT_LINECOM;
    default:
      return PT_TEXT;
    };
}



Parser::Parser(const char *_fname)
{
  fname=strdup(_fname);
  f=fopen(_fname,"r");
  lineno=1;
  if(f != NULL)
    gentoken();
  else
    {
      perror(0,ERROR_FILE_ROPEN,"Cannot open input file '%s': %s",_fname,
	     strerror(errno));
      curtok=PT_EOF;
    };
}

Parser::~Parser(void)
{
  if(f!=NULL)
    fclose(f);

  if(fname!=NULL)
    free(fname);
}



void Parser::error(int errnum, const char *args, ...)
{
  va_list ap;

  if(ERROR_IS_WARNING(errnum) && (msglevel<1))
    return;

  if(fname!=NULL)
    printf("%s, ",fname);
  printf("line %d: ",lineno);
  if(errnum>0)
    {
      if(ERROR_IS_WARNING(errnum))
	printf("Warning: ");
      else
	printf("Error: ");
    }
  va_start(ap, args);  
  vprintf(args, ap);
  va_end(ap);
  printf("\n");

  if(errnum > 0)
    describe_error(errnum);
}
